#include <WiFi.h>

// NTP settings
#define TIMEZONE 9 // timezone (GMT = 0, Japan = 9)
#define NTP_SERVER "pool.ntp.org"

#define WIFI_SMARTCONFIG true

#if !WIFI_SMARTCONFIG
// if you do not use smartConfifg, please specify SSID and password here
#define WIFI_SSID "SSID" // your WiFi's SSID
#define WIFI_PASS "PASS" // your WiFi's password
#endif

//////////////////////////////
// motor control section
//////////////////////////////

// Motor parameters
#define STEPS_PER_ROTATION 2048 // steps of a single rotation of motor
#define MOTOR_DELAY 2400
#define INITIAL_POS 18 // compensating backlash

#define PINS 4
// ports used to control the stepper motor.
//int port[PINS] = {27, 25, 32, 12};
int port[PINS] = {12, 4, 0, 2};

#define PHASE 4
// sequence of stepper motor control
int seq[PHASE][PINS] = {
  {  LOW,  LOW, HIGH,  LOW},
  {  LOW,  LOW,  LOW, HIGH},
  { HIGH,  LOW,  LOW,  LOW},
  {  LOW, HIGH,  LOW,  LOW}
};

// stepper motor rotation
void rotate(long step) {
  static int phase = 0;
  int delta = (step > 0) ? 1 : (PHASE-1);
  
  step = (step > 0) ? step : -step;
  for(long j = 0; j < step; j++) {
    phase = (phase + delta) % PHASE;
    for(int i = 0; i < PINS; i++) {
      digitalWrite(port[i], seq[phase][i]);
    }
    delayMicroseconds(MOTOR_DELAY);
    if(j < 20) delayMicroseconds(MOTOR_DELAY); // acceleration
    if(j < 10) delayMicroseconds(MOTOR_DELAY); // acceleration
    if(step - j < 10) delayMicroseconds(MOTOR_DELAY); // deacceleration
    if(step - j < 5) delayMicroseconds(MOTOR_DELAY); // deacceleration
    if(step - j < 2) delayMicroseconds(MOTOR_DELAY); // deacceleration
  }
  // power cut
  for(int i = 0; i < PINS; i++) {
    digitalWrite(port[i], LOW);
  }
}

// avoid error accumuration of fractional part of 4096 / POSITION
void rotStep(long s) {
  static long currentPos;
  
  long diff = s - currentPos;
  rotate(diff);
  currentPos = s;
}

//////////////////////////////
// WiFi and NTP section
//////////////////////////////

void getNTP(void) {
  for(int i = 0; WiFi.status() != WL_CONNECTED; i++) {
    if(i > 30) {
      ESP.restart();
    }
    Serial.println("Waiting for WiFi connection..");
    delay(1000);
  }

  configTime(TIMEZONE * 3600L, 0, NTP_SERVER);
  printLocalTime();
}

void printLocalTime() {
  struct tm timeinfo;
  if (!getLocalTime(&timeinfo)) {
    Serial.println("Failed to obtain time");
    return;
  }
  Serial.println(&timeinfo, "%A, %Y-%m-%d %H:%M:%S");
}

void wifiSetup() {
  int wifiMotion = STEPS_PER_ROTATION / 10; // while wainting for wifi, large motion
  int smatconfigMotion = STEPS_PER_ROTATION / 50; // while wainting for smartConfig, small motion
  int smartconfigCount = 0;

  WiFi.mode(WIFI_STA);
#if WIFI_SMARTCONFIG
  WiFi.begin();
#else
  WiFi.begin(WIFI_SSID, WIFI_PASS);
#endif

  for (int i = 0; ; i++) {
    Serial.println("Connecting to WiFi...");
    rotate(wifiMotion); wifiMotion *= -1;
    delay(1000);
    if (WiFi.status() == WL_CONNECTED) {
      break;
    }
#if WIFI_SMARTCONFIG
  if(i > 6)
    break;
#endif    
  }

#if WIFI_SMARTCONFIG
  if (WiFi.status() != WL_CONNECTED) {
    WiFi.mode(WIFI_AP_STA);
    WiFi.beginSmartConfig();

    //Wait for SmartConfig packet from mobile
    Serial.println("Waiting for SmartConfig.");
    while (!WiFi.smartConfigDone()) {
      Serial.print(".");
      rotate(smatconfigMotion); smatconfigMotion *= -1;
      delay(1000);
      smartconfigCount++;
      if(smartconfigCount > 600) { // wait for 10 min, then reset
        ESP.restart();
      }
    }

    Serial.println("");
    Serial.println("SmartConfig received.");

    //Wait for WiFi to connect to AP
    Serial.println("Waiting for WiFi");
    while (WiFi.status() != WL_CONNECTED) {
      rotate(wifiMotion); wifiMotion *= -1;
      delay(1000);
      Serial.print(",");
    }
  }
  Serial.println("WiFi Connected.");
#endif

  Serial.print("SSID: ");
  Serial.println(WiFi.SSID());
  Serial.print("IP Address: ");
  Serial.println(WiFi.localIP());
}

//////////////////////////////
// setup and main loop
//////////////////////////////

void setup() {
  Serial.begin(115200);
  Serial.println("start");

  pinMode(port[0], OUTPUT);
  pinMode(port[1], OUTPUT);
  pinMode(port[2], OUTPUT);
  pinMode(port[3], OUTPUT);

  wifiSetup();
  
  rotate(-STEPS_PER_ROTATION * 12); // reset using mechanical endstop
  
  rotate(INITIAL_POS);
  
  getNTP(); // get current time
}

void loop() {
  struct tm tmtime;
  long pos;

  getLocalTime(&tmtime);
  
  if(tmtime.tm_hour == 23 && tmtime.tm_min == 59 && tmtime.tm_sec >= 30) {
    ESP.restart(); // reset everyday
  }
  
  pos = (STEPS_PER_ROTATION * (tmtime.tm_min + tmtime.tm_hour * 60)) / 60;
  rotStep      (pos);
}
